from typing import List

import typer

from hometaxbot.crypto import load_cert, open_files
from hometaxbot.scraper import HometaxScraper

app = typer.Typer()


@app.command()
def certinfo(cert: List[str], cert_password):
    with open_files(cert) as files:
        sign = load_cert(files, cert_password)
        print('cert_class:', sign.cert_class(), '/ cn:', sign.cn(), '/ serialnum:', sign.serialnum())


scrape = typer.Typer()


@app.command()
def register_cert(registration_no, cert: List[str] = None, cert_password=None):
    scraper = HometaxScraper()
    res = scraper.register_cert(registration_no, cert, cert_password)
    print(res)


if __name__ == "__main__":
    app()


from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from io import BytesIO
from typing import List

from hometaxbot.types import HometaxModel


class 홈택스사용자구분코드(Enum):
    개인 = '01'
    개인사업자 = '02'
    법인사업자 = '03'


class 세목코드(Enum):
    종합소득세 = '10'
    원천세 = '14'
    양도소득세 = '22'
    법인세 = '31'
    부가세 = '41'


@dataclass(kw_only=True)
class 홈택스사용자(HometaxModel):
    납세자번호: str
    납세자명: str
    홈택스ID: str
    홈택스사용자번호: str
    사용자구분: 홈택스사용자구분코드
    휴대전화번호: str
    전자메일주소: str


@dataclass(kw_only=True)
class 납세자(HometaxModel):
    납세자번호: str
    납세자명: str
    휴대전화번호: str = None
    전자메일주소: str = None
    주소: str = None
    사업자구분: 홈택스사용자구분코드 = None

    대표자주민등록번호: str = None
    법인등록번호: str = None
    대표자명: str = None
    업종코드: str = None
    업태: str = None
    종목: str = None
    개업일: date = None
    폐업일: date = None
    사업장소재지: str = None
    사업장전화번호: str = None
    간이과세여부: bool = False


@dataclass(kw_only=True)
class 세무대리인:
    상호: str
    사업자번호: str
    관리번호: str
    사업장소재지: str = None
    대표자성명: str = None
    대표자주민등록번호: str = None
    전화번호: str = None


@dataclass(kw_only=True)
class 세무대리수임정보:
    납세자: 납세자
    세무대리인: 세무대리인
    수임일: date
    정보제공범위: str


@dataclass(kw_only=True)
class 전자신고결과조회(HometaxModel):
    접수번호: str
    세무서코드: str = None
    세목코드: 세목코드
    신고서종류: str # 예시: 원천징수이행상황신고서
    신고서종류코드: str = None  # 예시: F01
    신고구분코드: str = None  # 예시: 01
    신고구분상세코드: str = None  # 예시: 01
    신고구분: str  # 예시: 정기(확정)
    신고유형: str  # 예시: 정기신고
    과세연월: str
    과세기간시작일: date = None  # 예시: 20191101
    과세기간종료일: date = None  # 예시: 20191130
    지급연월: str = None
    접수일: date
    접수일시: datetime = None  # 예시: 20191203145924
    접수여부_첨부서류: str = None  # 예시: 정상(1종)
    홈택스ID: str = None
    납세자명: str = None
    납세자번호: str = None
    업종코드: str = None
    공급가액: Decimal  # 예시: 10500000
    금액: Decimal
    납부금액: Decimal


@dataclass(kw_only=True)
class 납부내역(HometaxModel):
    세무서코드: str
    세무서명: str
    납부일: date
    금액: Decimal
    전자납부발행번호: str
    세목: str
    세목코드: str
    결정구분: str = None
    신고연월: str = None
    전자납부번호: str = None

    def __post_init__(self):
        if not self.전자납부번호:
            self.전자납부번호 = '0126' + self.신고연월[2:] + self.결정구분 + self.세목코드 + self.전자납부발행번호


@dataclass(kw_only=True)
class 환급금조회(HometaxModel):
    환급번호: str
    세무서코드: str
    세무서명: str
    세목: str
    세목코드: str
    환급결정일: date
    최종변경일시: datetime
    귀속연월: date
    금액: Decimal


class 세금계산서분류(Enum):
    세금계산서 = '01'
    수정세금계산서 = '02'
    계산서 = '03'
    수정계산서 = '04'


class 세금계산서종류(Enum):
    일반 = '01'
    영세율 = '02'
    위수탁 = '03'
    수입 = '04'
    영세율위수탁 = '05'
    수입납부유예 = '06'


@dataclass(kw_only=True)
class 세금계산서(HometaxModel):
    승인번호: str
    전송일자: date  # 세금계산서 XML에는 포함되어 있지 않는 필드로 홈택스 시스템에서만 제공한다.
    작성일자: date
    세금계산서분류: 세금계산서분류
    세금계산서종류: 세금계산서종류
    영수청구코드: str
    수정코드: str = None
    당초승인번호: str = None
    비고: str = None
    수입문서참조: '수입문서' = None

    공급자: 납세자
    공급자연락처: '연락처'
    공급받는자: 납세자
    공급받는자연락처: '연락처'
    공급받는자연락처2: '연락처' = None
    위수탁자: 납세자 = None
    위수탁자연락처: '연락처' = None

    결제방법코드: str
    결제금액: Decimal
    공급가액: Decimal
    세액: Decimal
    총금액: Decimal

    품목: List['세금계산서품목']


@dataclass(kw_only=True)
class 수입문서(HometaxModel):
    신고번호: str
    일괄발급시작일: date
    일괄발급종료일: date
    총건: int


@dataclass(kw_only=True)
class 연락처(HometaxModel):
    부서명: str
    이름: str
    전화번호: str
    이메일: str


@dataclass(kw_only=True)
class 세금계산서품목(HometaxModel):
    일련번호: int
    공급일자: date
    품목명: str
    규격: str = None
    비고: str = None
    수량: Decimal = None # `-` 허용. 단위는 attribute로 들어간다.
    단가: Decimal  # 소수점 2자리까지 표현. `-` 허용
    공급가액: Decimal  # 원단위까지. `-` 허용
    세액: Decimal


@dataclass(kw_only=True)
class 카드매입(HometaxModel):
    거래일시: datetime
    카드번호: str
    승인번호: str
    카드사: str
    공급가액: Decimal
    부가세: Decimal
    봉사료: Decimal
    총금액: Decimal
    가맹점: 납세자
    가맹점유형: str
    공제여부: str
    비고: str


@dataclass(kw_only=True)
class 현금영수증(HometaxModel):
    거래일시: datetime
    매출매입: str
    승인번호: str
    승인구분: str
    발행구분: str = None
    발급수단: str
    거래구분: str = None
    공급가액: Decimal
    부가세: Decimal
    봉사료: Decimal
    총금액: Decimal
    매입자명: str = None
    가맹점: 납세자
    공제여부: bool = None


@dataclass(kw_only=True)
class 카드매출월간집계(HometaxModel):
    거래연월: date
    거래건수: int
    합계금액: Decimal
    매입처명: str


@dataclass(kw_only=True)
class 체납내역:
    잔여납부세액: Decimal  # 3470280
    납세자번호: str  # 129-86-67290
    과세기간_세목: str  # 2018년2기분 부가가치세
    감액세액: Decimal  # 0
    납세자명: str  # 주식회사 이콜레모
    고지세액: Decimal   # 3119410
    일부납부세액: Decimal  # 0
    전자납부번호: str   # 0126-1903-5-41-11168940
    납부URL: str = None
    상태: str  # 미납
    세목명: str  # 부가가치세
    세목코드: str  # 41


@dataclass(kw_only=True)
class 고지내역:
    세목코드: str  # 41
    전자납부번호: str  # 0126-2003-5-41-22057360
    고지번호: str   # 100000000047718675
    납세자번호: str  # 129-86-67290
    납부URL: str = None
    고지세액: Decimal  # 710540
    납세자명: str  # 주식회사 이콜레모
    납부기한: date  # 2020-03-31
    납부의무종결사유: str = None  # 해당없음
    고지송달일: date
    납부여부: str  # 미납
    고지서유형: str  # 일반고지
    분납_연부연납세액: Decimal  # 0
    납부할세액: Decimal  # 710540
    과세기간_세목: str  # 2019년2기분 부가가치세
    감액세액: Decimal  # 0
    세목: str  # 부가가치세


@dataclass(kw_only=True)
class 전자신고_신고서_납부서(HometaxModel):
    납세자: 납세자
    # 신고내역: 전자신고결과조회
    신고서_data: dict
    납부내역: 납부내역
    납부서_pdf: BytesIO



import csv
import json
import time
from datetime import date
from decimal import Decimal
from io import StringIO

import dateutil.parser
from dateutil.relativedelta import relativedelta

from hometaxbot import models, Throttled
from hometaxbot.models import 수입문서, 납세자, 연락처, 세금계산서품목
from hometaxbot.scraper import HometaxScraper, parse_response
from hometaxbot.scraper.requestutil import XMLValueFinder, action_params, get_quarter_by_date
from hometaxbot.scraper.util import split_date_range

invoice_type_choices = {
    '전자세금계산서': '01',
    '전자세금계산서_위수탁': '02',
    '전자계산서': '03',
    '전자계산서_위수탁': '04',
}


def 세금계산서(scraper: HometaxScraper, begin: date, end: date):
    scraper.request_permission('teet')
    for invoice_type in invoice_type_choices:
        for 매입매출 in ['매입', '매출']:
            위수탁 = '위수탁' in invoice_type
            prhSlsClCd = "01" if 매입매출 == "매출" else "02"
            if 위수탁:
                prhSlsClCd = "02"

            조회기준코드 = {
                '작성일자': '01',
                '발급일자': '02',
                '전송일자': '03',
            }

            for period_begin, period_end in split_date_range(begin, end, relativedelta(months=3)):
                params = {
                    "bmanCd": "00",
                    "dmnrMpbNo": "",
                    "dmnrTxprDscmNo": "",
                    "dtCl": 조회기준코드["전송일자"],
                    "etxivClsfCd": "all",
                    "etxivKndCd": "all",
                    "inqrDtEnd": period_end.strftime("%Y%m%d"),
                    "inqrDtStrt": period_begin.strftime("%Y%m%d"),
                    "isnTypeCd": "all",
                    "pageNum": "",
                    "pageSize": "10",
                    "prhSlsClCd": prhSlsClCd,
                    "screenId": "",
                    "splrMpbNo": "",
                    "splrTxprDscmNo": "",
                    "tnmNm": "",
                    "cstnBmanMpbNo": "",
                    "cstnBmanTin": scraper.tin if 위수탁 else "",
                    "dmnrTin": scraper.tin if 매입매출 == "매입" and not 위수탁 else "",
                    "dmnrTnmNm": "",
                    "etxivClCd": invoice_type_choices[invoice_type],
                    "gubunCd": "",
                    "mCd": "",
                    "mqCd": "",
                    "qCd": "",
                    "splrTin": scraper.tin if 매입매출 == "매출" and not 위수탁 else "",
                    "splrTnmNm": "",
                    "tmsnDtIn": "",
                    "tmsnDtOut": "",
                    "yCd": ""
                }
                first_page = scraper.request_action_json("ATEETBDA001R01", 'UTEETBDA01', json={
                    "cstnInfoYn": "",
                    "fleDwldYn": "",
                    "fleTp": "",
                    "icldCstnBmanInfr": "",
                    "icldLsatInfr": "N",
                    "resnoSecYn": "Y",
                    "srtClCd": "1",
                    "srtOpt": "02",
                    "etxivIsnBrkdTermDVOPrmt": params
                }, subdomain='teet')

                if first_page['pageInfoVO']['totalCount'] == 0:
                    continue

                for page in range(1, first_page['pageInfoVO']['totalCount'] // scraper.DOWNLOAD_PAGE_SIZE + 2):
                    pageInfoVO = {
                        'pageSize': scraper.DOWNLOAD_PAGE_SIZE,
                        'pageNum': page,
                        'totalCount': first_page['pageInfoVO']['totalCount']
                    }
                    res = scraper.session.post('https://teet.hometax.go.kr/wqAction.do', data={
                        'downloadParam': json.dumps({
                            "cstnInfoYn": "", "fleDwldYn": "Y", "fleTp": "txt", "icldCstnBmanInfr": "",
                            "icldLsatInfr": "", "resnoSecYn": "Y", "srtClCd": "1", "srtOpt": "01",
                            "affectedCnt": 0,
                            "pageInfoVO": pageInfoVO,
                            "excelPageInfoVO": pageInfoVO,
                            "etxivIsnBrkdTermDVOPrmt": params,
                        }),
                        'actionId': 'ATEETBDA005R04',
                        'screenId': 'UTEETBDA01',
                        'noopen': False,
                        'downloadView': 'Y'
                    })

                    reader = StringIO(res.content.decode('utf8'))
                    for i in range(4):
                        next(reader)

                    dict_reader = csv.DictReader(reader, delimiter='\t')
                    for row in dict_reader:
                        yield scrape_세금계산서_detail(scraper, row['승인번호'].replace('-', ''), row['전송일자'])
                        time.sleep(Throttled.wait)


def scrape_세금계산서_detail(scraper: HometaxScraper, etan, 전송일자):
    """전송일자는 세금계산서 XML 내에 없고 홈택스 시스템에서 관리하는 값이기 때문에 외부에서 전달 받아야 한다."""
    scraper.request_permission('teet')
    etan = etan.replace('-', '')
    res = scraper.session.post("https://teet.hometax.go.kr/wqAction.do",
                               data={
                                   'downloadParam': json.dumps({
                                       "fileDwnYn": "Y",
                                       "etan": etan,
                                       "etxivIsnBrkdTermDVOPrmt": {
                                           "etan": etan,
                                           "screenId": "UTEETBDA01",
                                           "slsPrhClCd": "01",
                                           "etxivClCd": "",
                                           "etxivClsfCd": "",
                                           "etxivMpbNo": "0",
                                           "etxivTin": scraper.tin,
                                           "pageNum": 1,
                                           "focus": "resultGrid_cell_0_11",
                                           "layerPopup": "Y",
                                           "callbackFn": "mf_txppWframe___close_callback_Func__1730730392626_2",
                                           "__popupName": "전자세금계산서 상세조회 팝업",
                                           "popupID": "UTEETBDA38"
                                       }
                                   }).replace(' ', ''),
                                   'actionId': 'ATEETBDA001R02',
                                   'screenId': 'UTEETBDA38',
                                   "downloadView": "Y",
                                   "noopen": False,
                               },
                               headers={'Content-Type': "application/x-www-form-urlencoded"})

    res.encoding = 'UTF-8'
    finder = XMLValueFinder(parse_response(res))

    return models.세금계산서(
        승인번호=finder.get('TaxInvoiceDocument/IssueID'),
        전송일자=전송일자,
        작성일자=finder.get('TaxInvoiceDocument/IssueDateTime'),
        세금계산서분류=finder.get('TaxInvoiceDocument/TypeCode')[:2],
        세금계산서종류=finder.get('TaxInvoiceDocument/TypeCode')[2:],
        영수청구코드=finder.get('TaxInvoiceDocument/PurposeCode'),
        수정코드=finder.get('TaxInvoiceDocument/AmendmentStatusCode'),
        당초승인번호=finder.get('TaxInvoiceDocument/OriginalIssueID'),
        비고=finder.get('TaxInvoiceDocument/DescriptionText'),
        수입문서참조=finder.get('TaxInvoiceDocument/ReferencedImportDocument') and 수입문서(
            신고번호=finder.get('TaxInvoiceDocument/ReferencedImportDocument/ID'),
            일괄발급시작일=finder.get('TaxInvoiceDocument/ReferencedImportDocument/AccetablePeriod/StartDateTime'),
            일괄발급종료일=finder.get('TaxInvoiceDocument/ReferencedImportDocument/AccetablePeriod/EndDateTime'),
            총건=finder.get('TaxInvoiceDocument/ReferencedImportDocument/ItemQuantity')),
        공급자=납세자(
            납세자번호=finder.get('TaxInvoiceTradeSettlement/InvoicerParty/ID'),
            납세자명=finder.get('TaxInvoiceTradeSettlement/InvoicerParty/NameText'),
            대표자명=finder.get('TaxInvoiceTradeSettlement/InvoicerParty/SpecifiedPerson/NameText'),
            주소=finder.get('TaxInvoiceTradeSettlement/InvoicerParty/SpecifiedAddress/LineOneText'),
            업태=finder.get('TaxInvoiceTradeSettlement/InvoicerParty/TypeCode'),
            종목=finder.get('TaxInvoiceTradeSettlement/InvoicerParty/ClassificationCode'),
            업종코드=finder.get('TaxInvoiceTradeSettlement/InvoicerParty/ClassificationCode')),
        공급자연락처=연락처(
            부서명=finder.get('TaxInvoiceTradeSettlement/InvoicerParty/DefinedContact/DepartmentNameText'),
            이름=finder.get('TaxInvoiceTradeSettlement/InvoicerParty/DefinedContact/PersonNameText'),
            전화번호=finder.get('TaxInvoiceTradeSettlement/InvoicerParty/DefinedContact/TelephoneCommunication'),
            이메일=finder.get('TaxInvoiceTradeSettlement/InvoicerParty/DefinedContact/URICommunication')),
        공급받는자=납세자(
            납세자번호=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/ID'),
            납세자명=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/NameText'),
            대표자명=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/SpecifiedPerson/NameText'),
            주소=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/SpecifiedAddress/LineOneText'),
            업태=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/TypeCode'),
            종목=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/ClassificationCode'),
            업종코드=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/ClassificationCode')),
        공급받는자연락처=연락처(
            부서명=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/PrimaryDefinedContact/DepartmentNameText'),
            이름=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/PrimaryDefinedContact/PersonNameText'),
            전화번호=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/PrimaryDefinedContact/TelephoneCommunication'),
            이메일=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/PrimaryDefinedContact/URICommunication')),
        공급받는자연락처2=연락처(
            부서명=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/SecondaryDefinedContact/DepartmentNameText'),
            이름=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/SecondaryDefinedContact/PersonNameText'),
            전화번호=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/SecondaryDefinedContact/TelephoneCommunication'),
            이메일=finder.get('TaxInvoiceTradeSettlement/InvoiceeParty/SecondaryDefinedContact/URICommunication')),
        위수탁자=납세자(
            납세자번호=finder.get('TaxInvoiceTradeSettlement/BrokerParty/ID'),
            납세자명=finder.get('TaxInvoiceTradeSettlement/BrokerParty/NameText'),
            대표자명=finder.get('TaxInvoiceTradeSettlement/BrokerParty/SpecifiedPerson/NameText'),
            주소=finder.get('TaxInvoiceTradeSettlement/BrokerParty/SpecifiedAddress/LineOneText'),
            업태=finder.get('TaxInvoiceTradeSettlement/BrokerParty/TypeCode'),
            종목=finder.get('TaxInvoiceTradeSettlement/BrokerParty/ClassificationCode'),
            업종코드=finder.get('TaxInvoiceTradeSettlement/BrokerParty/ClassificationCode')),
        위수탁자연락처=연락처(
            부서명=finder.get('TaxInvoiceTradeSettlement/BrokerParty/SecondaryDefinedContact/DepartmentNameText'),
            이름=finder.get('TaxInvoiceTradeSettlement/BrokerParty/SecondaryDefinedContact/PersonNameText'),
            전화번호=finder.get('TaxInvoiceTradeSettlement/BrokerParty/SecondaryDefinedContact/TelephoneCommunication'),
            이메일=finder.get('TaxInvoiceTradeSettlement/BrokerParty/SecondaryDefinedContact/URICommunication')),
        결제방법코드=finder.get('TaxInvoiceTradeSettlement/SpecifiedPaymentMeans/TypeCode'),
        결제금액=finder.get('TaxInvoiceTradeSettlement/SpecifiedPaymentMeans/PaidAmount'),
        공급가액=finder.get('TaxInvoiceTradeSettlement/SpecifiedMonetarySummation/ChargeTotalAmount'),
        세액=finder.get('TaxInvoiceTradeSettlement/SpecifiedMonetarySummation/TaxTotalAmount'),
        총금액=finder.get('TaxInvoiceTradeSettlement/SpecifiedMonetarySummation/GrandTotalAmount'),
        품목=[세금계산서품목(
            일련번호=f.get('SequenceNumeric'),
            공급일자=f.get('PurchaseExpiryDateTime'),
            품목명=f.get('NameText'),
            규격=f.get('InformationText'),
            비고=f.get('DescriptionText'),
            수량=f.get('ChargeableUnitQuantity'),
            단가=f.get('UnitPrice/UnitAmount'),
            공급가액=f.get('InvoiceAmount'),
            세액=f.get('TotalTax/CalculatedAmount'),
        ) for f in finder.sub_finders('TaxInvoiceTradeLineItem')]
    )


def 카드매입(scraper: HometaxScraper, begin: date, end: date):
    scraper.request_permission('tecr')

    for period_begin, period_end in split_date_range(begin, end, relativedelta(months=3)):
        for element in scraper.paginate_action_json(
                'ATECRCCA001R06', 'UTECRCB023',
                json={
                    "busnCrdcDwldFleStatCd": "",
                    "busnCrdcTrsBrkdPrhYr": "",
                    "dwldCnclFg": "",
                    "dwldFleNm": "",
                    "dwldTrsBrkdScnt": "",
                    "fleTp": "",
                    "gdncMsgCntn": "",
                    "ntplBmanAthYn": "",
                    "prhQrt": "",
                    "prhQrtEdInq": "",
                    "prhQrtStInq": "",
                    "prhQrtStrtYm": "",
                    "prhTxamtDdcYn": "all",
                    "reqCd": "",
                    "resultCd": "",
                    "rqsDt": "",
                    "rqstTxprDscmNo": "",
                    "sumTotaTrsAmt": "",
                    "tin": scraper.tin,
                    "trsDtRngEnd": period_end.strftime("%Y%m%d"),
                    "trsDtRngStrt": period_begin.strftime("%Y%m%d"),
                    "txprDclsCd": "250",
                    "upldFleNm": "",
                    "upldPsbYn": "",
                    "yearInq": "",
                },
                subdomain='tecr'):

            yield models.카드매입(
                거래일시=element['aprvDt'],
                카드번호=element['busnCrdCardNoEncCntn'],
                승인번호=element['busnCrdcTrsBrkdSn'],
                카드사=element['crcmClNm'],
                공급가액=element['splCft'],
                부가세=element['vaTxamt'],
                봉사료=element['tip'],
                총금액=element['totaTrsAmt'],
                가맹점=납세자(
                    납세자번호=element['mrntTxprDscmNoEncCntn'],
                    납세자명=element['mrntTxprNm'],
                    업종코드=element['tfbCd'],
                    업태=element['bcNm'],
                    종목=element['tfbNm'],
                ),
                가맹점유형=element['bmanClNm'],
                공제여부=element['ddcYnNm'],
                비고=element['vatDdcClNm'],
            )


def 현금영수증(scraper: HometaxScraper, begin: date, end: date):
    scraper.request_permission('tecr')

    # 현금영수증 매출
    # 매입과 달리 매출의 경우만 body를 암호화하고 있어서 일반 요청 대신 다운로드 요청으로 전환한다.
    # 처음에는 일반 요청을 해서 몇몇 정보를 받은 후 그 정보를 이용해서 다시 다운로드 요청을 한다.
    for period_begin, period_end in split_date_range(begin, end, relativedelta(months=3)):
        first_page = scraper.session.post('https://tecr.hometax.go.kr/wqAction.do', params={
            "actionId": 'ATECRCBA001R03',
            "screenId": 'UTECRCB013',
            "popupYn": "false",
            "realScreenId": ''
        }, json={
            "dprtUserYn": "N",
            "fleTp": "",
            "mrntTxprDscmNoEncCntn": "",
            "pubcUserNo": "all",
            "reqCd": "",
            "spjbTrsYn": "all",
            "spstCnfrId": "all",
            "sumTotaTrsAmt": "",
            "tin": scraper.tin,
            "trsDtRngEnd": period_end.strftime("%Y%m%d"),
            "trsDtRngStrt": period_begin.strftime("%Y%m%d"),
            "txprDscmNo": scraper.selected_trader.납세자번호,
            "totalCount": "0",
            "sumSplCft": "0",
        }).json()

        if first_page['pageInfoVO']['totalCount'] == 0:
            continue

        DOWNLOAD_PAGE_SIZE = 1000
        for page in range(1, first_page['pageInfoVO']['totalCount'] // DOWNLOAD_PAGE_SIZE + 2):
            res = scraper.session.post('https://tecr.hometax.go.kr/wqAction.do', data={
                "downloadParam": json.dumps({
                    "fleTp": "txt",
                    "pblClCd": "all",
                    "reqCd": "00",
                    "sumTotaTrsAmt": first_page['sumTotaTrsAmt'],
                    "tin": scraper.tin,
                    "trsDtRngEnd": period_end.strftime("%Y%m%d"),
                    "trsDtRngStrt": period_begin.strftime("%Y%m%d"),
                    "totalCount": 0,
                    "sumSplCft": first_page['sumSplCft'],
                    "pageInfoVO": {
                        'pageSize': DOWNLOAD_PAGE_SIZE,
                        'pageNum': page,
                        'totalCount': first_page['pageInfoVO']['totalCount']
                    },
                }),
                'actionId': 'ATECRCBA001R03',
                'screenId': 'UTECRCB013',
                'noopen': False,
                'downloadView': 'Y',
            })

            reader = StringIO(res.content.decode('utf8'))
            next(reader)  # 첫 줄에 요약 정보 건너뛰기
            reader = csv.DictReader(reader, delimiter='\t')
            for row in reader:
                yield models.현금영수증(
                    매출매입='매출',
                    거래일시=dateutil.parser.parse(row['매출일시']),
                    거래구분=row['비고'],
                    공급가액=Decimal(row['공급가액'].replace(',', '')),
                    부가세=Decimal(row['부가세'].replace(',', '')),
                    봉사료=Decimal(row['봉사료'].replace(',', '')),
                    총금액=Decimal(row['총금액'].replace(',', '')),
                    승인번호=row['승인번호'],
                    발급수단='',
                    발행구분=row['발행구분'],
                    승인구분=row['거래구분'],
                    매입자명=row['신분확인뒷4자리'],
                    가맹점=scraper.selected_trader,
                )

    # 현금영수증 매입
    for element in scraper.paginate_action_json(
            'ATECRCBA001R02', 'UTECRCB005',
            json={
                "dprtUserYn": "N",
                "fleTp": "",
                "mrntTxprDscmNoEncCntn": "",
                "pubcUserNo": "all",
                "reqCd": "",
                "spjbTrsYn": "all",
                "spstCnfrId": "all",
                "sumTotaTrsAmt": "",
                "tin": scraper.tin,
                "trsDtRngEnd": end.strftime("%Y%m%d"),
                "trsDtRngStrt": begin.strftime("%Y%m%d"),
                "txprDscmNo": scraper.selected_trader.납세자번호,
                "totalCount": "0",
                "sumSplCft": "0",
            }, subdomain='tecr'):
        finder = XMLValueFinder(element)
        yield models.현금영수증(
            매출매입='매입',
            거래일시=element['trsDtTime'],
            거래구분=element['cshptTrsTypeNm'],
            공급가액=element['splCft'],
            부가세=element['vaTxamt'],
            봉사료=element['tip'],
            총금액=element['totaTrsAmt'],
            승인번호=element['aprvNo'],
            발급수단=element['spstCnfrClNm'],
            발행구분=element.get('pblClCd'),
            승인구분=element['trsClNm'],
            매입자명=element['rcprTxprNm'],
            가맹점=납세자(
                납세자번호=element['mrntTxprDscmNoEncCntn'],
                납세자명=element['mrntTxprNm'],
                업종코드=element['tfbCd'],
                업태=element['tfbNm'],
                종목=element['itmNm'],
            ),
            공제여부=element['prhTxamtDdcClNm'] == '공제' or element['prhTxamtDdcYn'] == 'Y',
        )


def 카드매출월간집계(scraper: HometaxScraper, begin: date, end: date):
    """
    자료수집 시점에 대한 홈택스의 안내문구:
        신용카드 자료는 매월 15일경에 직전월 자료까지 포함하여 제공합니다
        현금IC카드 자료는 부가세 신고월에 직전분기 자료까지 포함하여 제공합니다
        1~3월 매출자료 :  4월 10일경 제공      4~6월 매출자료 : 7월 10일경 제공
        7~9월 매출자료 :  10월 10일경 제공   10~12월 매출자료 : 1월 10일경 제공
        판매(결제)대행 매출자료는 부가가치법 제75조에 따라 매 분기 다음 달 15일까지 제출하도록 규정되어 있습니다
        따라서, 자료수집 후 신속하게 제공할 예정입니다
        1~3월 매출자료 :  4월 17일경 제공 예정      4~6월 매출자료 : 7월 17일경 제공 예정
        7~9월 매출자료 :  10월 17일경 제공 예정   10~12월 매출자료 : 1월 17일경 제공 예정
    """
    scraper.request_permission('teht')

    element = scraper.request_xml(
        'https://teht.hometax.go.kr/wqAction.do?' + action_params('ATESFAAA014R02', 'UTESFABG35'),
        payload=f'<map id="ATESFAAA014R02">'
                f'<stlYr>{begin.year}</stlYr>'
                f'<qrtFrom>{get_quarter_by_date(begin)}</qrtFrom>'
                f'<qrtTo>{get_quarter_by_date(end)}</qrtTo>'
                f'<bsno>{scraper.selected_trader.납세자번호}</bsno>'
                f'</map>')

    for child in element.findall('.//list/map'):
        finder = XMLValueFinder(child)
        yield models.카드매출월간집계(
            거래연월=finder.get('stlYm') + '01',  # 연월 값에 01을 붙여서 날짜로 만듦
            거래건수=finder.get('sumStlScnt'),
            합계금액=finder.get('sumTipExclAmt'),
            매입처명=finder.get('txprNm'),
        )




"""
홈택스에 세무대리인으로 로그인했을 때 보이는 메뉴에서 스크래핑하는 기능들.
"""
from datetime import date

from hometaxbot.scraper import HometaxScraper


def 수임납세자(scraper: HometaxScraper, begin: date, end: date):
    scraper.session.get('https://hometax.go.kr/websquare/websquare.html?w2xPath=/ui/pp/index_pp.xml&tmIdx=48&tm2lIdx=4804000000&tm3lIdx=4804040000')

    scraper.request_permission(screen_id='index_pp')

    res = scraper.request_action_json('ATXPPCBA001R17', 'index_pp', json={
        "cncClCd": "01",
        "srvcClCd": "01",
        "menuHtrnId": "4800000000",
        "ntplAthYn": "N",
        "ntplBmanAthYn": "N",
        "crpBmanAthYn": "Y",
        "txaaYn": "Y",
        "cshptMrntYn": "",
        "pubcUserNo": scraper.pubcUserNo,
        "dprtUserYn": "N",
        "athCd": "Y",
        "menuId": "",
        "prevMenuId": "",
        "menuTtl": ""
    })
    res = scraper.session.post('https://hometax.go.kr/userAthEvtxMenuUtil', data={'type': 6})
    res = scraper.request_action_json('ATXPPCBA001R019', 'index_pp', json={
        'scrnId': "4804040000", 'pageInfoVO': {'totalCount': "0", 'pageSize': "10", 'pageNum': "1"}
    }, nts_postfix=False)
    res = scraper.request_action_json('ATXPPCBA001R020', 'index_pp', json={
        'scrnId': "4804040000", 'pageInfoVO': {'totalCount': "0", 'pageSize': "10", 'pageNum': "1"}
    }, nts_postfix=False)
    res = scraper.request_action_json('ATXPPAAA001R037', 'index_pp', json={"ttxppal032DVO":{"menuId":""}})

    scraper.request_permission('teht', 'UTEABHAA03')
    for item in scraper.paginate_action_json('ATEABHAA001R10', 'UTEABHAA03', json={
        "afaDtChnfTrgtYn": "",
        "afaEndDt": end.strftime("%Y%m%d"),
        "afaStrtDt": begin.strftime("%Y%m%d"),
        "afdsCl": "1",
        "excelPageNum": "",
        "excelPageSize": "",
        "excelYn": "N",
        "flag": "",
        "txaaAdmNo": scraper.cta_admin_no,
        "txprDscmNoEncCntn": ""
    }, subdomain='teht'):
        yield item



import base64
import json
import logging
import os
import re
import ssl
import subprocess
import time
from datetime import datetime
from http import HTTPStatus
from typing import List, TypedDict, TypeVar
from urllib.parse import unquote_plus, unquote
from xml.etree import ElementTree
from xml.etree.ElementTree import Element

import requests
import yaml
from cryptography.hazmat.primitives._serialization import Encoding
from requests import JSONDecodeError

from hometaxbot import HometaxException
from hometaxbot import random_second, AuthenticationFailed, Throttled
from hometaxbot.crypto import load_cert, open_files, validate_cert_expiry, k4, snake_oil_encrypt
from hometaxbot.models import 홈택스사용자구분코드, 홈택스사용자, 납세자, 세무대리인, 세무대리수임정보
from hometaxbot.scraper.requestutil import nts_generate_random_string, ensure_xml_response, parse_response, \
    check_error_on_response, CustomHttpAdapter, json_minified_dumps


class HometaxScraper:
    LOGIN_SUCCESS_CODE = 'S'
    HOMETAX_REQUEST_TIMEOUT = 7
    PAGE_SIZE = 10
    DOWNLOAD_PAGE_SIZE = 1000

    user_info: 홈택스사용자 = None
    selected_trader: 납세자 = None
    개인사업자_list: List[TypedDict('개인사업자', {'사업자등록번호': str, 'tin': str})] = []
    tin: str = None
    user_tin: str = None
    pubcUserNo: str = None
    subdomain: str = None

    def __init__(self):
        ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        ctx.options |= 0x4  # OP_LEGACY_SERVER_CONNECT
        self.session = requests.session()
        self.session.mount('https://', CustomHttpAdapter(ctx))
        self.session.headers.update({
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.0.0 Safari/537.36',
        })
        # self.session.headers['Content-Type'] = 'application/json'
        self.cta_admin_no = None

    def login_with_cert(self, cert_paths: List[str], prikey_password):
        """
        홈택스 공인인증서 로그인
        """
        nts = random_second()
        self.session.post('https://hometax.go.kr/wqAction.do?actionId=ATXPPCBA001R17&screenId=index_pp&popupYn=false&realScreenId=',
                          data='{"cncClCd":"01","srvcClCd":"01","menuHtrnId":"100900","ntplAthYn":"N","ntplBmanAthYn":"","crpBmanAthYn":"","txaaYn":"","cshptMrntYn":"","pubcUserNo":"","dprtUserYn":"","athCd":"N","menuId":"","prevMenuId":"","menuTtl":""}<nts<nts>nts>59h3OdIFTyGfSN5JWA5ycXWy2NySYP0nOT5DAy4zN848')
        self.request_permission(screen_id='UTXPPABA01')
        self.session.post('https://hometax.go.kr/jsp/magicNX/getMLSession.jsp', data='')
        self.session.post('https://hometax.go.kr/wqAction.do?actionId=ATXPPABA001A25&screenId=UTXPPABA01&popupYn=false&realScreenId=', data='{"ipUData":""}<nts<nts>nts>60sSHhJ0npixTDedljXKTgyCTViu40xfLWs6aimQmq8A49')
        self.session.post('https://hometax.go.kr/wqAction.do?actionId=ATXPPCBA001R020&screenId=UTXPPABA01&popupYn=false', json={"scrnId": "0900000000", "pageInfoVO": {"totalCount" : "0", "pageSize" : "10", "pageNum" : "1"} })

        res = self.session.post("https://www.hometax.go.kr/wqAction.do?actionId=ATXPPZXA001R01&screenId=UTXPPABA01",
                                data=f"{{}}{nts}dGGBLG2rRWBeuYMviAZyJjAphI9Y3wCmWhg1y84EU{nts - 11}",
                                headers={'Content-Type': 'application/json'})
        if res.status_code == HTTPStatus.NOT_FOUND:
            raise HometaxException('홈택스 서버에 접속할 수 없습니다. 잠시 후 다시 시도해주세요.')
        pckEncSsn = res.json()['pkcEncSsn']

        with open_files(cert_paths) as files:
            sign = load_cert(files, prikey_password)
            validate_cert_expiry(sign)

        if len(cert_paths) == 1:
            p = subprocess.Popen(['openssl', 'pkcs12', '-info', '-provider', 'legacy', '-provider', 'default',
                                  '-in', cert_paths[0], '-nodes', '-nocerts', '-passin',
                                  f'pass:{prikey_password}'], stdout=subprocess.PIPE)
            prikey_dumped, _ = p.communicate()
            ID_KISA_NPKI_RAND_NUM = '1.2.410.200004.10.1.1.3'
            result = re.search(f'{ID_KISA_NPKI_RAND_NUM}: (.*)\n-----BEGIN PRIVATE KEY-----', prikey_dumped.decode())
            rand_num = result.group(1)
            randomEnc = base64.b64encode(bytearray.fromhex(rand_num)).decode('utf8')
        else:
            randomEnc = base64.b64encode(sign._rand_num.asOctets()).decode('utf-8')

        content = f'{pckEncSsn}$' \
                  f'{str(hex(sign.serialnum())).replace("x", "")}$' \
                  f'{datetime.now().strftime("%Y%m%d%H%M%S")}$' \
                  f'{base64.b64encode(sign.sign(pckEncSsn.encode("utf-8"))).decode("utf-8")}'
        logSgnt = base64.b64encode(content.encode('utf-8')).decode('utf-8')
        cert = sign.pub_cert.public_bytes(Encoding.PEM).decode('utf-8')

        res = self.session.post("https://www.hometax.go.kr/pubcLogin.do",
                                params={"domain": "hometax.go.kr", "mainSys": "Y"},
                                data={
                                    "logSgnt": logSgnt,
                                    "cert": cert,
                                    "randomEnc": randomEnc,
                                    "pkcLoginYnImpv": 'Y',
                                    "pkcLgnClCd": '03',
                                    "scrnId": 'UTXPPABA01&',
                                    "userScrnRslnXcCnt": 1680,
                                    "userScrnRslnYcCnt": 1050
                                }, headers={"content-type": "application/x-www-form-urlencoded; charset=UTF-8"})

        result_code_match = re.search(r"'code' : '(.)'", res.text)
        if result_code_match is None:
            raise HometaxException(f'홈택스 서버 오류입니다. 홈택스 응답: {res.text}')

        result_code = result_code_match.group(1)
        if result_code != self.LOGIN_SUCCESS_CODE:
            err_msg = re.search(r"'errMsg' : decodeURIComponent\('(.+?)'\)", res.text).group(1)
            unquoted = unquote_plus(err_msg).replace('+', ' ').replace('\\n', '\n')
            auth_errors = ['홈택스에 등록된 인증서가 아닙니다', '폐지된 인증서입니다']
            if len([e for e in auth_errors if e in unquoted]):
                raise AuthenticationFailed(re.sub(r'\[[^\]]+\]', '', unquoted))
            else:
                logging.exception(res.text)
                raise Exception(unquoted)

        self.fetch_user_and_traders()

    def register_cert(self, registration_no, cert_paths, prikey_password):
        """
        공동인증서를 홈택스에 등록한다.
        """
        self.request_permission(screen_id='UTXPPABA14')

        nts = random_second()
        res = self.request_action(
            action_id='ATXPPABA003R02',
            screen_id='UTXPPABA14',
            payload=json.dumps({
                "crpBmanPkcRgtNbSu": "",
                "crpPkcRgtVrfYn": "",
                "ntplBmanCrpBmanCl": "",
                "tin": "",
                "txprDscmNo": registration_no}
            ) + f'<nts<nts>nts>{nts}mAGVLUUQsmZ8Uv9YJj9bAtbDAQARJTjesj6HaT{nts - 11}').json()

        res = self.request_action(
            'ATXPPZXA001R01',
            'UTXPPABA14',
            payload=f"{{}}<nts<nts>nts>{nts}dGGBLG2rRWBeuYMviAZyJjAphI9Y3wCmWhg1y84EU{nts - 11}").json()
        pckEncSsn = res['pkcEncSsn']

        with open_files(cert_paths) as files:
            sign = load_cert(files, prikey_password)

        if len(cert_paths) == 1:
            p = subprocess.Popen(['openssl', 'pkcs12', '-info', '-provider', 'legacy', '-provider', 'default',
                                  '-in', cert_paths[0], '-nodes', '-nocerts', '-passin',
                                  f'pass:{prikey_password}'], stdout=subprocess.PIPE)
            prikey_dumped, _ = p.communicate()
            ID_KISA_NPKI_RAND_NUM = '1.2.410.200004.10.1.1.3'
            result = re.search(f'{ID_KISA_NPKI_RAND_NUM}: (.*)\n-----BEGIN PRIVATE KEY-----', prikey_dumped.decode())
            rand_num = result.group(1)
            randomEnc = base64.b64encode(bytearray.fromhex(rand_num)).decode('utf8')
        else:
            randomEnc = base64.b64encode(sign._rand_num.asOctets()).decode('utf-8')

        content = f'{pckEncSsn}$' \
                  f'{str(hex(sign.serialnum())).replace("x", "")}$' \
                  f'{datetime.now().strftime("%Y%m%d%H%M%S")}$' \
                  f'{base64.b64encode(sign.sign(pckEncSsn.encode("utf-8"))).decode("utf-8")}'
        logSgnt = base64.b64encode(content.encode('utf-8')).decode('utf-8')
        cert = sign.pub_cert.public_bytes(Encoding.PEM).decode('utf-8')

        res = self.request_action(
            action_id='ATXPPABA001C01',
            screen_id='UTXPPABA14',
            payload=json.dumps({
                'cert': cert,
                'hashCntn': '',
                'logSgnt': logSgnt,
                'pkcCertClCd': '',
                'pkcEncSsn': '',
                'randomEnc': randomEnc,
                'skey': '',
                'tin': '',
                'txprDscmNo': registration_no,
            }) + f'</map><nts<nts>nts>{nts}b8FxcemzOmtdbg6smXGQTS9XHdH665qRZQSiW1XfXTg{nts - 11}').json()
        if res['resultMsg']['result'] != self.LOGIN_SUCCESS_CODE:
            if res['resultMsg']['msg'] in ['선택하신 인증서는 이미 등록된 인증서 입니다.']:
                return res['resultMsg']['msg']
            raise Exception(res['resultMsg']['msg'])
        return res['resultMsg']['msg']

    def fetch_user_and_traders(self):
        self.deselect_trader()

        res_json = self.session.post('https://www.hometax.go.kr/permission.do',
                                params={"screenId": "index"},
                                data='<map id="postParam"><popupYn>false</popupYn></map>'.encode('utf-8'),
                                headers={'Content-Type': "application/xml; charset=UTF-8"}).json()
        self.check_authenticated(res_json)
        pubcUserNo = res_json['resultMsg']['sessionMap']['pubcUserNo']
        nts = random_second()
        res_json = self.request_action(
            action_id='ATXPPAAA001R22',
            screen_id='UTXPPAAA10',
            payload=json.dumps({"pubcUserNo": pubcUserNo, "userType": "B", "cncClCd": "", "arsPswdAltYn": "", "jntCnt": ""})
                    + f'<nts<nts>nts>{nts}BfI2b32na00UC4Gq5TCUjAlsw7uURISbBokVb2ShBc0{nts - 11}').json()

        self.user_info = model_from_hometax_json(홈택스사용자, res_json['pubcUserJnngInfrAdmDVO'])
        # self.user_tin = res_json.find('.//map[@id="sessionMap"]/tin').text
        if self.user_info.사용자구분 == 홈택스사용자구분코드.개인:
            self.개인사업자_list = self.fetch_traders()
        else:
            self.selected_trader = self.trader_info()

    def fetch_traders(self):
        nts = random_second()
        res_json = self.request_action(
            action_id='ATXPPAAA003R01',
            screen_id='UTXPPAAA24',
            payload=f'{{}}'
                    f'<nts<nts>nts>{nts}NJ3QcOLdNy8YZIojAqeUQiS6YP653gruVRI9JbNVw{nts - 11}').json()
        elements = res_json.get('bmanBscInfrInqrDVOList')

        if elements is None:
            if res_json.find('.//msg') is None or '사업자 변경대상이 아님' not in res_json.find('.//msg').text:
                logging.error(ElementTree.tostring(res_json, encoding='unicode'), stack_info=True)
            return []

        return [{
            '사업자등록번호': element['txprDscmNoEncCntn'],
            'tin': element['tin'],
        } for element in elements]

    def login_with_cookies(self, cookies):
        """
        짧은 시간 안에 여러 번 스크래핑할 때는 반복적으로 공동인증서 로그인을 사용하기보다 한 번 로그인하고 그 쿠키를 재활용한다.
        """
        for cookie in cookies:
            if 'httpOnly' in cookie:
                http_only = cookie.pop('httpOnly')
                cookie['rest'] = {'httpOnly': http_only}
            if 'expiry' in cookie:
                cookie['expires'] = cookie.pop('expiry')
            self.session.cookies.set(**cookie)

        self.fetch_user_and_traders()

    def cookies(self):
        COOKIE_ATTRS = ["version", "name", "value", "port", "domain", "path", "secure", "expires", "discard", "comment",
                        "comment_url", "rfc2109"]
        return [{attr: getattr(cookie, attr) for attr in COOKIE_ATTRS} for cookie in self.session.cookies]

    def request_action_xml(self, action_id, screen_id, real_screen_id='', payload: str = None):
        return ensure_xml_response(self.request_action,
                                   action_id,
                                   screen_id,
                                   real_screen_id=real_screen_id,
                                   payload=payload.encode('utf8'),
                                   content_type='application/xml; charset=UTF-8')

    def request_action_json(self, action_id, screen_id, json: dict, real_screen_id='', popup_yn='false', subdomain: str = None, nts_postfix=True):
        # TODO 응답 오류 처리
        # 오류 예시: {'resultMsg': {'detailMsg': '', 'msg': '6개월이상은 조회할 수 없습니다. \n 조회에 실패 실패하였습니다.', 'exceptType': 'APPLICATION', 'serviceTxId': 'PTEET1103_ATEETBDA001R01_T00799_1731908564438', 'detailLogYn': 'N', 'code': 'ETICMZ0008', 'result': 'F'}}
        res = self.session.post(f'https://{subdomain + '.' if subdomain else ''}hometax.go.kr/wqAction.do',
                                params={"actionId": action_id, "screenId": screen_id, "popupYn": popup_yn,
                                         "realScreenId": real_screen_id},
                                data=self.nts_postfix_added(json) if nts_postfix else json_minified_dumps(json),
                                headers={'Content-Type': 'application/json'}, timeout=20)
        try:
            data = res.json()
            if data['resultMsg']['result'] != self.LOGIN_SUCCESS_CODE:
                raise HometaxException(data['resultMsg']['detailMsg'] or data['resultMsg']['msg'])
            return data
        except JSONDecodeError as e:
            raise HometaxException(f'홈택스 응답오류: {res.text}')


    def paginate_action_json(self, action_id, screen_id, json: dict, real_screen_id='', subdomain: str = None,
                             page_begin=1, page_end=None):
        page = page_begin
        while True:
            pageInfoVO = {
                'pageNum': page, 'pageSize': self.PAGE_SIZE, 'totalCount': 0
            }
            data = self.request_action_json(action_id, screen_id,
                                            json | {'pageInfoVO': pageInfoVO},
                                            real_screen_id, subdomain=subdomain)
            try:
                yield from data[next(k for k in data if k.endswith('VOList'))]
            except StopIteration:
                return

            if 'pageInfoVO' not in data:
                return # 페이지 정보가 없으면 더 이상 페이지가 없다고 가정한다.

            if data['pageInfoVO']['totalCount'] is None:
                raise Exception(f'홈택스 응답 오류: {data}')

            if page_end and page >= page_end:
                return

            if page * self.PAGE_SIZE > data['pageInfoVO']['totalCount']:
                return

            page += 1
            # 홈택스 쓰로틀링에 걸리는 걸 방지하기 위해 페이지마다 간격을 둔다.
            time.sleep(Throttled.wait)

    def request_action(self, action_id, screen_id, real_screen_id='', payload=None, content_type='application/xml; charset=UTF-8'):
        return self.session.post('https://hometax.go.kr/wqAction.do',
                                 params={
                                     "actionId": action_id,
                                     "screenId": screen_id,
                                     "popupYn": "false",
                                     "realScreenId": real_screen_id
                                 },
                                 data=payload,
                                 headers={'Content-Type': content_type}, timeout=20)

    def select_trader(self, 사업자등록번호):
        if self.user_info.사용자구분 != 홈택스사용자구분코드.개인:
            return

        try:
            found = next(filter(lambda trader: trader['사업자등록번호'] == 사업자등록번호, self.개인사업자_list))
        except StopIteration:
            raise AuthenticationFailed(f'홈택스에서 사업자등록번호 "${사업자등록번호}"인 사업자를 찾을 수 없습니다.')

        nts = random_second()
        self.request_permission(screen_id='UTXPPAAA24')
        res = self.request_action(action_id='ATXPPAAA003A01',
                            screen_id='UTXPPAAA24',
                            payload=json.dumps({'tin': found['tin']}) +
                                    f'<nts<nts>nts>{nts}YOp3ShJZFIhX5xqYRB0ELzlkd0EehhPVkHF6mjLJk{nts - 11}')
        self.tin = None
        self.request_permission(screen_id='index')
        if self.tin != found['tin']:
            logging.exception(f'사업자 선택에 실패했습니다. user_tin : {self.user_tin} target: {found["tin"]} result: {self.tin}')
            raise Throttled(60 * 5, '사업자 선택에 실패했습니다. 다시 시도해주세요.')

        self.selected_trader = self.trader_info()

    def deselect_trader(self):
        """
        개인 인증서로 로그인한 경우 사업자가 선택되어 있는 상태에서 개인이 선택된 상태로 변경한다.
        개인사업자나 법인 인증서의 경우는 홈택스 응답에서 에러가 나지만 무시한다.
        :return:
        """
        nts = random_second()
        self.request_permission(screen_id='UTXPPAAA24')
        res = self.request_action(action_id='ATXPPAAA003A01',
                                  screen_id='UTXPPAAA24',
                                  payload=f'<map id="ATXPPAAA003A01"><tin>ORIGIN</tin></map>'
                                          f'<nts<nts>nts>{nts}YOp3ShJZFIhX5xqYRB0ELzlkd0EehhPVkHF6mjLJk{nts - 11}', )
        self.selected_trader = None

    def trader_info(self, debug=False):
        self.request_permission('teht', 'UTEABGAA21')
        nts = random_second()
        res = self.session.post(
            "https://teht.hometax.go.kr/wqAction.do?actionId=ATTABZAA001R17&screenId=UTEABGAA21&popupYn=false&realScreenId=",
            data=json_minified_dumps({"tin": self.tin, "txprClsfCd": '02', "txprDscmNo": "", "txprDscmNoClCd": "",
                                      "txprDscmDt": "", "searchOrder": "02/01", "outDes": "bmanBscInfrInqrDVO",
                                      "txprNm": "", "crpTin": "", "mntgTxprIcldYn": "", "resnoAltHstrInqrYn": "",
                                      "resnoAltHstrInqrBaseDtm": "", "sameBmanInqrYn": "N", "rpnBmanRetrYn": "N"})
                 + f'<nts<nts>nts>{nts}rjNjDYrX04H1ZeoLR7s39xGAggSKKn7ZTGjjfyMK0{nts - 11}',
            headers={'Content-Type': "application/json"})

        element = res.json()['bmanBscInfrInqrDVO']
        if element is None:
            logging.error(f'trader_info none element: {res.text}', extra=dict(text=res.text), stack_info=True)
            raise Exception('사업자 정보를 불러오지 못했습니다. 다시 시도해주세요.')

        self.request_permission(screen_id='UTXPPBAA69')
        nts = random_second()
        # 마이홈택스 - 기타 세무정보 - 사업자등록사항 및 담당자 안내
        res = self.session.post(
            "https://hometax.go.kr/wqAction.do?actionId=ATXPPBAA001R36&screenId=UTXPPBAA70&popupYn=true&realScreenId=",
            data=json.dumps({"rprsTin": "", "tin": self.tin})
                 + f'</map><nts<nts>nts>{nts}74JxrC2hphMsLLv7deT0nri5fT4KO9iHHdSTK9SATM{nts - 11}',
            headers={'Content-Type': "application/xml; charset=UTF-8"})

        detail = res.json().get('bmanBscInfrInqrDVOList')
        if detail is None:
            raise Throttled('사업자 등록사항을 불러오지 못했습니다. 다시 시도해주세요.')

        사업자상태 = element.get('txprStatNm')
        사업자등록사항 = next(
            (d for d in detail if d['txprDscmNoEncCntn'].replace('-', '') == element['txprDscmNoEncCntn']),
            None)

        return 납세자(
            납세자번호=element['txprDscmNoEncCntn'],
            사업자구분=홈택스사용자구분코드[element['txprDclsNm']],
            전자메일주소=element['pfbEml'],
            휴대전화번호=element['pfbTelno'],
            주소=element['roadAdr'],
            납세자명=element['txprNm'] if element['txprNm'] else element['rprsTxprNm'],
            대표자주민등록번호=element['rprsResno'].replace('-', ''),
            법인등록번호=element.get('crpno'),
            대표자명=element['rprsTxprNm'],
            업종코드=element.get('tfbCd', element.get('xmtxOrgMtfbCd', 'ZZZZZZ')),
            업태=element.get('bcNm', element.get('xmtxOrgMbcNm')) if 사업자상태 != '폐업' else None,
            종목=element.get('itmNm', element.get('xmtxOrgMitmNm')) if 사업자상태 != '폐업' else None,
            개업일=datetime.strptime(element['txprDscmDt'], '%Y%m%d').date(),
            폐업일=element.get('cfbDt') and datetime.strptime(element['cfbDt'], '%Y%m%d').date(),
            사업장소재지=element['rprsRoadAdr'],
            사업장전화번호=element['rprsHmTelno'],
            간이과세여부='간이과세자' in 사업자등록사항['bmanClNm'] if 사업자등록사항 is not None else False,
        )

    def request_permission(self, subdomain=None, screen_id=None):
        """
        목록조회에 필요한 세션을 얻어옴 = > TEETsessionID
        subdomain:
            - teet: '세금계산서'
            - tecr: '현금영수증매출'
            - teht: '카드매출'
        :return:
        tin: 권한을 얻은 사업자의 홈택스 고유의 식별자
        """
        if self.subdomain == subdomain and self.tin and self.pubcUserNo:
            return

        if screen_id is None:
            screen_id = 'UTEETBAA99'

        base_url = 'https://'
        if subdomain is None:
            base_url += 'hometax.go.kr'
        else:
            base_url += f'{subdomain}.hometax.go.kr'

        response = self.session.post(f'{base_url}/permission.do?screenId={screen_id}', json={}, timeout=20)

        root = response.json()
        if subdomain and ('sessionMap' not in root['resultMsg']):
            token = self.session.post(f'https://hometax.go.kr/token.do?query=_{nts_generate_random_string(20)}', json={}).json()
            root = self.session.post(f'{base_url}/permission.do?screenId={screen_id}&domain=hometax.go.kr',
                                     json=token | {'popupYn': False,},
                                     timeout=20).json()

        if 'sessionMap' in root['resultMsg']:
            self.tin = root['resultMsg']['sessionMap']['tin']
            self.pubcUserNo = root['resultMsg']['sessionMap']['pubcUserNo']
            self.txprClsfCd = root['resultMsg']['sessionMap']['txprClsfCd']  # TODO 이 값을 요청할 때 쓰는 게 맞는지 확인 필요.
        elif subdomain:
            raise Throttled(60, '홈택스 로그인 권한 획득에 실패했습니다. 다시 시도해주세요.')

        self.subdomain = subdomain

    def request_xml(self, url, payload):
        return parse_response(self.session.post(
            url,
            data=self.nts_postfix_added(payload),
            headers={'Content-Type': "application/xml; charset=UTF-8"}
        ))

    def request_paginated_xml(self, url, payload: str):
        page = 1
        while True:
            res_xml = self.request_xml(url, payload.format(
                pageInfoVO=f'<map id="pageInfoVO">'
                           f'<pageSize>{self.PAGE_SIZE}</pageSize><pageNum>{page}</pageNum>'
                           '<totalCount>0</totalCount>'
                           f'</map>'))

            yield from res_xml.findall('.//list/map')

            if res_xml.find('.//map[@id="pageInfoVO"]/totalCount') is None:
                raise Exception('홈택스 응답 오류: ' + ElementTree.tostring(res_xml, encoding='utf-8').decode('utf-8'))

            if page * self.PAGE_SIZE > int(res_xml.find('.//map[@id="pageInfoVO"]/totalCount').text):
                return

            page += 1
            # 홈택스 쓰로틀링에 걸리는 걸 방지하기 위해 페이지마다 간격을 둔다.
            time.sleep(Throttled.wait)

    def nts_postfix_added(self, data: json):
        second = datetime.now().strftime('%0S')
        payload = json_minified_dumps(data)
        return payload + f'<nts<nts>nts>{int(second) + 11}{k4(payload, second, userId=self.user_info.홈택스ID)}{second}'

    def check_authenticated(self, res_json: dict):
        if 'resultMsg' not in res_json:
            raise Exception('홈택스 응답 오류: ' + json.dumps(res_json, ensure_ascii=False))

        if 'sessionMap' not in res_json['resultMsg']:
            raise AuthenticationFailed('홈택스에 로그인되지 않은 상태입니다.')

    def fetch_세무대리수임정보(self):
        data = self.request_action_json('ATXPPBAA001R18', 'UTXPPBAB61', {})
        for item in data['txaaInfrDVOList']:
            yield 세무대리수임정보(
                납세자=납세자(납세자번호=item['txprDscmNoEncCntn'], 납세자명=item['txprTnm']),
                세무대리인=세무대리인(
                    상호=item['txaaAdmTnmNm'],
                    사업자번호=item['bsno'],
                    전화번호=item['txaaAdmTelno'],
                    관리번호=item['txaaAdmNoEncCntn']
                ),
                정보제공범위=item['infrOfrRngCdNm'],
                수임일=item['afaDt'],
            )

    def 사업자등록상태(self):
        data = self.request_action_json('ATTABZAA001R08', 'UTEABAAA13', {
            "dongCode": self.selected_trader.납세자번호[3:5],
            "inqrTrgtClCd": "1",
            "mobYn": "N",
            "psbSearch": "Y",
            "pubcUserNo": self.pubcUserNo,
            "txprDscmNo": self.selected_trader.납세자번호,
            "userReqInfoVO": {}
        })
        result = {
            '등록상태': data['trtCntn']
        }
        result['면세구분'] = '면세' if any([keyword in result['등록상태'] for keyword in ['면세', '고유번호가 부여된 단체']]) else '과세'
        if '전환된 날짜' in result['등록상태']:
            try:
                result['과세유형전환'] = dict(
                    전환일자=datetime.strptime(re.search(r'전환된 날짜는 (.+?) 입니다.', result['등록상태']).group(1), '%Y년 %m월 %d일').date(),
                    전환유형=re.search(r'부가가치세 (.+?) 입니다.', result['등록상태']).group(1)
                )
            except Exception as e:
                logging.error(e)
        return result

    def login_as_tax_accountant(self, ctn_no, cta_password):
        res = self.request_action_json('ATXPPCBA001R17', 'index_pp', json={
            "cncClCd": "01", "srvcClCd": "01", "menuHtrnId": "100900", "ntplAthYn": "N", "ntplBmanAthYn": "N",
            "crpBmanAthYn": "Y", "txaaYn": "Y", "cshptMrntYn": "", "pubcUserNo": self.pubcUserNo,
            "dprtUserYn": "N", "athCd": "Y", "menuId": "", "prevMenuId": "", "menuTtl": ""})

        self.request_permission(screen_id='UTEABHAA19')

        res = self.request_action_json('ATXPPAAA006R43', 'UTEABHAA19', json={"ttxppal032DVO":{"menuId":""}})

        res = self.session.post('https://hometax.go.kr/pubcLogin.do?operate=txaaLogin', data={
            'txaaAdmNo': ctn_no,
            'txaaPswd': snake_oil_encrypt(cta_password),
        })

        login_res = self.parse_cta_login_response(res.text)
        postfix_str = datetime.today().strftime("%Y_%m_%d")
        res = self.session.get(
            f'https://teht.hometax.go.kr/pubcLogin.do?operate=txaaSession&txaaAdmNo={login_res[1]["lgnRsltCd"]}&postfix={postfix_str}')

        self.request_permission(screen_id='index_pp')
        res = self.session.post(
            'https://hometax.go.kr/wqAction.do?actionId=ATXPPCBA001R17&screenId=index_pp&popupYn=false&realScreenId=',
            data=json.dumps({
                "cncClCd": "01",
                "srvcClCd": "01",
                "menuHtrnId": "1100000000",
                "ntplAthYn": "N",
                "ntplBmanAthYn": "N",
                "crpBmanAthYn": "Y",
                "txaaYn": "Y",
                "cshptMrntYn": "",
                "pubcUserNo": self.pubcUserNo,
                "dprtUserYn": "N",
                "athCd": "Y",
                "menuId": "",
                "prevMenuId": "",
                "menuTtl": ""}) + '<nts<nts>nts>32k8yZmB2hWqWIwyDqt5v50oHY58M4RpVphs78jPDZg21',
            headers={'Content-Type': 'application/json'}
        )
        if res.json()['resultMsg']['result'] != self.LOGIN_SUCCESS_CODE:
            raise AuthenticationFailed('홈택스에 로그인되지 않은 상태입니다.')

        self.cta_admin_no = ctn_no

    def parse_cta_login_response(self, data: str) -> tuple[str, dict, str]:
        callback_pattern = r'(?<=nts_loginSystemCallback\()(.*?)(?=\);)'
        data1 = re.search(callback_pattern, data).group(0)
        data1 = re.sub(r'(?=\'errMsg\')(.*?)(?=\'lgnRsltCd\')', '', data1)
        data1 = data1.replace("'", '"')

        decode_pattern = r'(?<=decodeURIComponent\()(.*?)(?=\);)'
        data2 = re.search(decode_pattern, data).group(0)
        data2 = re.sub(r'(?=\'lgnRsltCd\')(.*?)(?=\'pswdErrNbcnt\')', '', data2)
        data2 = data2.replace("'", '"')

        sys_code = re.search(r'(?<=").*?(?=")', data1).group(0)

        json_str = re.search(r'(?<=\,)(.*?)(?<=\})', data1).group(0)
        parsed_json = json.loads(json_str)

        err_msg = re.search(r'(?<=").*?(?=")', data2).group(0)
        err_msg = unquote(err_msg).replace('+', ' ').replace('\\n', '\n')

        return sys_code, parsed_json, err_msg


with open(os.path.dirname(__file__) + '/hometax_xml_fields.yml', encoding='utf-8') as f:
    HOMETAX_XML_FIELDS = yaml.safe_load(f)


def model_from_hometax_xml(model_class, element: Element):
    return model_class(**{HOMETAX_XML_FIELDS[model_class.__name__][child.tag]: child.text
                          for child in element if HOMETAX_XML_FIELDS[model_class.__name__].get(child.tag)})


T = TypeVar("T")


def model_from_hometax_json(model_class: T, data: dict) -> T:
    return model_class(**{HOMETAX_XML_FIELDS[model_class.__name__][key]: value
                          for key, value in data.items() if HOMETAX_XML_FIELDS[model_class.__name__].get(key)})


def find_first_value(element, *names, default=''):
    for name in names:
        if element.find(name) is not None and element.find(name).text:
            return element.find(name).text

    return default



import base64
import logging
import re
import time
from pathlib import Path

from selenium import webdriver
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.wait import WebDriverWait
from webdriver_manager.chrome import ChromeDriverManager

from hometaxbot import HometaxException

WAIT_LONG = 20
WAIT_SHORT = 3


class HometaxDriver:
    def __init__(self, download_folder=None, headless=True):
        options = webdriver.ChromeOptions()
        if headless:
            options.add_argument('--headless')

        options.add_argument('--no-sandbox')  # https://github.com/theintern/intern/issues/878
        options.add_argument(
            '--disable-dev-shm-usage')  # https://developers.google.com/web/tools/puppeteer/troubleshooting - Tips

        self.download_folder = download_folder or str(Path.home() / 'Downloads')

        preferences = {"download.default_directory": self.download_folder,
                       "directory_upgrade": True,
                       "safebrowsing.enabled": True}
        options.add_experimental_option("prefs", preferences)

        self.driver = webdriver.Chrome(service=Service(ChromeDriverManager().install()), options=options)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.driver.close()
        self.driver.quit()

    def login_with_cert(self, cert_path, password):
        self.home()
        self.wait_and_click(By.LINK_TEXT, '로그인')

        self.wait_loading()
        WebDriverWait(self.driver, 10).until(lambda locator: self.driver.execute_script('return Boolean(window.WebSquare.uiplugin.editor)'))
        while True:
            try:
                WebDriverWait(self.driver, WAIT_SHORT).until_not(
                    expected_conditions.presence_of_element_located((By.CLASS_NAME, 'blockUI')))
                self.wait_and_click(By.XPATH, './/i[text()[contains(., "공동·금융인증서")]]/../..')
                self.wait(By.ID, 'dscert', delay=WAIT_SHORT)
                self.driver.switch_to.frame('dscert')

                WebDriverWait(self.driver, WAIT_SHORT).until(
                    expected_conditions.presence_of_element_located((By.CLASS_NAME, 'blockUI')))
                WebDriverWait(self.driver, WAIT_LONG).until_not(
                    expected_conditions.presence_of_element_located((By.CLASS_NAME, 'blockUI')))
                self.wait_and_click(By.ID, 'in_browser')
                break
            except TimeoutException as e:
                print('retry...', e)
                self.driver.switch_to.default_content()
                pass

        self.wait_for_blockui_disappeared()
        self.driver.find_element(By.ID, 'filefile2').send_keys(cert_path)
        # self.driver.find_element(By.ID, 'add_browser_password').send_keys(password)
        self.driver.execute_script(f'document.getElementById("add_browser_password").value = "{password}"')

        try:
            self.driver.execute_script('document.getElementById("add_browser_password_layout").style.display = "none"')
        except:
            pass

        self.driver.find_element(By.CSS_SELECTOR, '.n_blue_btn #btn_common_confirm').click()
        try:
            self.wait_and_click(By.CLASS_NAME, 'occui_bt_close', WAIT_SHORT)
        except:
            pass

        self.driver.switch_to.default_content()

    def login_as_tax_agent(self, ctn_no, cta_password):
        for element in self.driver.find_elements(By.CLASS_NAME, 'w2textbox'):
            if "세무대리인 권한을 가진 사용자" in element.text:
                self.wait_and_click(By.XPATH, '//input[@value="확인"]')
                self.wait(By.XPATH, '//input[@title="세무대리인 관리번호 입력"]')
                (self.driver.find_element(By.XPATH, '//input[@title="세무대리인 관리번호 입력"]')
                 .send_keys(ctn_no))
                (self.driver.find_element(By.XPATH, '//input[@title="비밀번호 입력"]')
                 .send_keys(cta_password))
                self.wait_and_click(By.XPATH, '//input[@value="로그인"]')
                break
        else:
            raise HometaxException('세무대리인 인증서가 아닙니다.')

    def wait_for_blockui_disappeared(self, delay=WAIT_LONG):
        WebDriverWait(self.driver, delay).until(
            expected_conditions.presence_of_element_located((By.CLASS_NAME, 'blockUI')))
        WebDriverWait(self.driver, delay).until_not(
            expected_conditions.presence_of_element_located((By.CLASS_NAME, 'blockUI')))

    def login(self, username, password):
        self.home()
        self.wait_and_click(By.LINK_TEXT, '로그인')

        WebDriverWait(self.driver, WAIT_LONG).until(lambda locator: self.driver.execute_script('return !!window.jQuery && window.jQuery.active == 0'))
        self.switch_to_txpp_frame()
        self.wait_and_click(By.LINK_TEXT, '아이디 로그인')
        self.driver.execute_script(self.javascript_for_login(username, password))

        self.driver.get('https://hometax.go.kr')
        self.wait(By.LINK_TEXT, '로그아웃')

        # self.wait_and_click(By.CSS_SELECTOR, '.login_state a:first-child')
        # self.wait(By.NAME, 'txppIframe')
        #
        # self.driver.switch_to.frame('txppIframe')
        # self.wait(By.NAME, 'iptUserId').send_keys(username)
        # self.driver.find_element_by_xpath('//input[@name="iptUserPw"]').send_keys(password)
        # self.wait_loading()
        # self.driver.find_element_by_class_name('login_box').find_element_by_xpath('//input[@type="button"][@value="로그인"]').click()

    def home(self):
        # self.driver.get('https://hometax.go.kr')
        self.driver.get('https://hometax.go.kr/websquare/websquare_cdn.html?w2xPath=/ui/pp/index.xml')
        self.wait_loading()

    @staticmethod
    def javascript_for_login(username, password):
        return re.sub(r'\s', '', f"""
            window.$.post('/pubcLogin.do?domain=hometax.go.kr&mainSys=Y', {{
                ssoLoginYn: 'Y',
                secCardLoginYn: '',
                secCardId: '',
                cncClCd: '01',
                id: '{base64.b64encode(username.encode('utf-8')).decode('utf-8')}',
                pswd: '{password}',
                ssoStatus: '',
                portalStatus: '',
                scrnId: 'UTXPPABA01',
                userScrnRslnXcCnt: '1680',
                userScrnRslnYcCnt: '1050'
            }}).then(function(res) {{ 
                eval(res)
            }})
        """)

    def wait(self, by, value, delay=WAIT_LONG):
        WebDriverWait(self.driver, delay).until(expected_conditions.presence_of_element_located((by, value)))
        return self.driver.find_element(by, value)

    def wait_and_click(self, by, value, delay=WAIT_LONG):
        WebDriverWait(self.driver, delay).until(expected_conditions.element_to_be_clickable((by, value))).click()

    def wait_loading(self):
        WebDriverWait(self.driver, WAIT_LONG).until(expected_conditions.invisibility_of_element((By.CLASS_NAME, 'w2_modal')))
        WebDriverWait(self.driver, WAIT_LONG).until(lambda locator: self.driver.execute_script('return !!window.jQuery && window.jQuery.active == 0'))

    def close(self):
        self.driver.close()
        self.driver.quit()

    def account_info(self):
        self.wait_loading()
        self.wait_and_click(By.CSS_SELECTOR, '.user_info a.btn_info')
        self.wait_loading()
        self.switch_to_txpp_frame()

        info = {}
        WebDriverWait(self.driver, 10).until(expected_conditions.visibility_of_any_elements_located((By.CSS_SELECTOR, 'table.tbl_form')))
        tables = self.driver.find_elements(By.CSS_SELECTOR, 'table.tbl_form')
        if tables[0].is_displayed():
            # 개인
            table = tables[0]
        else:
            # 사업자
            table = tables[1]

        for tr in table.find_elements(By.TAG_NAME, 'tr'):
            for th, td in zip(tr.find_elements(By.TAG_NAME, 'th'), tr.find_elements(By.TAG_NAME, 'td')):
                info[th.text] = td.text

        table = self.driver.find_element(By.CSS_SELECTOR, 'table.tbl_form2')
        for tr in table.find_elements(By.TAG_NAME, 'tr'):
            for th, td in zip(tr.find_elements(By.TAG_NAME, 'th'), tr.find_elements(By.TAG_NAME, 'td')):
                info[th.text] = td.text.split('\n')[0]

        return info

    def switch_to_txpp_frame(self):
        self.driver.switch_to.default_content()
        self.driver.switch_to.frame('txppIframe')

    def upload_report(self, report_type, filename):
        self.wait_and_click(By.LINK_TEXT, '신고/납부')
        self.switch_to_txpp_frame()
        if report_type == '종합소득세':
            self.wait_and_click(By.LINK_TEXT, '종합소득세')
            # 100000000000710000 window close
            self.wait_and_click(By.LINK_TEXT, '일반 신고')
            self.wait_and_click(By.CSS_SELECTOR, '#grpt02l0501')  # 파일변환 신고 클릭
            self.accept_alert()
        elif report_type == '부가가치세':
            self.wait_and_click(By.LINK_TEXT, '부가가치세')
            self.wait_and_click(By.LINK_TEXT, '파일 변환신고\n(회계프로그램)')
        else:
            raise Exception(f'{report_type} is not supported')

        # file upload
        self.wait(By.ID, 'raonkuploader_frame_fileList')
        self.driver.switch_to.frame('raonkuploader_frame_fileList')
        self.driver.find_element(By.CSS_SELECTOR, 'input[type="file"]').send_keys(filename)
        self.switch_to_txpp_frame()
        self.driver.find_element(By.CSS_SELECTOR, "input[value='형식검증하기']").click()
        self.accept_alert()
        WebDriverWait(self.driver, WAIT_LONG).until(
            expected_conditions.presence_of_element_located((By.ID, 'UTERNAAZ65_iframe')))
        self.driver.switch_to.frame('UTERNAAZ65_iframe')
        self.driver.find_element(By.ID, 'input1').send_keys('12345678')
        self.driver.find_element(By.CSS_SELECTOR, 'input[value="확인"]').click()

        result = {
            'result': [],
            'errors': [],
            'wetax': False,
            'wetax_retry': 0
        }
        time.sleep(1)
        self.switch_to_txpp_frame()
        self.driver.find_element(By.CSS_SELECTOR, "input[value='형식검증결과확인']").click()

        if self.cell('#grid1_body_table', 1, 3).text != '0' and self.cell('#grid1_body_table', 2, 2).text != '0':
            self.cell('#grid1_body_table', 1, 3).find_element(By.TAG_NAME, 'a').click()
            result['errors'] = self.driver.find_element(By.ID, 'group2792_UTERNAAZ45').text.split('\n')
            self.wait_and_click(By.CSS_SELECTOR, 'input[value="이전"]')

        self.switch_to_txpp_frame()

        if self.cell('#grid1_body_table', 1, 4).text != '1':
            result['result'] = '형식검증오류'
            return result

        self.driver.find_element(By.CSS_SELECTOR, "input[value='내용검증하기']").click()
        time.sleep(3)
        self.driver.find_element(By.CSS_SELECTOR, "input[value='내용검증결과확인']").click()

        if self.cell('#grid1_body_table', 1, 5).text != '0' and self.cell('#grid1_body_table', 2, 3).text != '0':
            result['errors'] = list(self.report_validation_errors('#grid1_body_table', 5))

        if self.cell('#grid1_body_table', 1, 6).text != '1':
            result['result'] = '내용검증오류'
            return result

        self.switch_to_txpp_frame()
        self.driver.find_element(By.CSS_SELECTOR, "input[value='전자파일제출']").click()
        self.driver.find_element(By.CSS_SELECTOR, "input[type='radio']").click()
        self.driver.find_element(By.CSS_SELECTOR, "input[value='전자파일 제출하기']").click()
        self.accept_alert()
        self.accept_alert()

        self.driver.switch_to.frame('UTERNAAZ02_iframe')
        self.driver.find_element(By.CSS_SELECTOR, "input[value='닫기']").click()
        result['result'] = '신고완료'
        return result

    def send_report_to_wetax(self, 대표자주민등록번호, 전화번호):
        self.wait_and_click(By.LINK_TEXT, '신고/납부')
        self.switch_to_txpp_frame()
        self.wait_and_click(By.LINK_TEXT, '종합소득세')

        self.driver.find_element(By.LINK_TEXT, '신고내역 조회 (접수증 · 납부서)').click()
        self.driver.find_element(By.CSS_SELECTOR, 'input[value="조회하기"]').click()
        self.accept_alert()
        self.wait_and_click(By.CSS_SELECTOR, 'td#ttirnam101DVOListDes_cell_0_37')
        self.accept_alert()
        self.wait_and_click(By.ID, 'jumin')
        self.driver.find_element(By.ID, 'jumin').send_keys(대표자주민등록번호[-7:])
        self.wait_and_click(By.ID, 'btn_jumin')
        self.accept_alert()
        self.wait_and_click(By.ID, 'checkSingo')
        self.driver.find_element(By.ID, 'moTel').send_keys(전화번호)
        self.wait_and_click(By.ID, 'btn_regi')
        self.accept_alert()

        return True

    def report_validation_errors(self, table, col):
        # 현재 종합소득세만 동작
        self.cell(table, 1, col).find_element(By.TAG_NAME, 'a').click()
        self.wait_and_click(By.CSS_SELECTOR, 'table.gridHeaderTableDefault tr td a')
        self.driver.switch_to.window('UTERNAAZ13')
        for tr in self.driver.find_elements(By.CSS_SELECTOR, '#ErrBrkdInqrDVOListDes_body_table tr'):
            if not tr.find_elements(By.TAG_NAME, 'td'):
                continue

            print(dict(zip(['서식명', '메시지유형', '항목명', '내용검증 오류내역'], [td.text for td in tr.find_elements(By.TAG_NAME, 'td')])))

        self.driver.close()
        self.driver.switch_to.window(self.driver.window_handles[0])
        self.switch_to_txpp_frame()
        self.wait_and_click(By.CSS_SELECTOR, 'input[value="이전"]')

    def cell(self, table_selector, row, col):
        return self.driver.find_element(By.CSS_SELECTOR, f'{table_selector} tr:nth-child({row}) td:nth-child({col})')

    def accept_alert(self):
        try:
            WebDriverWait(self.driver, 3).until(expected_conditions.alert_is_present(), 'timeout')

            alert = self.driver.switch_to.alert
            alert.accept()

        except TimeoutException:
            logging.warning("no alert")



import json
import random
import re
import time
from xml.etree import ElementTree

import requests.adapters
import urllib3

from requests import Response

from hometaxbot import Throttled, HometaxException


def action_params(action, screen):
    return f"actionId={action}&screenId={screen}&popupYn=false&realScreenId="


def nts_generate_random_string(length):
    """
    홈택스 스크립트에 있는 랜덤 스트링 제너레이션 포팅
    :param length: 생성할 스트링 길이
    :return:
    """
    seed = "qwertyuiopasdfghjklzxxcvbnm0123456789QWERTYUIOPASDDFGHJKLZXCVBNBM"
    result = ''
    for i in range(length):
        result += seed[random.randint(0, len(seed) - 1)]
    return result


def check_error_on_response(response):
    if '요청하신 서비스는 현재 서비스 중지 시간 입니다' in response.text:
        raise Throttled(response.text.strip())

    if '반복적인 호출은 시스템에 부하를 줄 수 있습니다' in response.text:
        raise Throttled(response.text.strip())

    if '조회에 실패하였습니다' in response.text:
        raise Throttled(response.text.strip())

    if '서비스 실행 중 오류가 발생하였습니다.' in response.text:
        raise HometaxException(response.text)

    if response.status_code == 400:
        if 'Request Blocked' in response.text:
            raise Throttled(response.text.strip())
        raise HometaxException(response.text)


def parse_response(response):
    """
    :param response: requests.Response
    :return: validated된 reponse를 xml 파싱하여 ElementTree객체로 리턴
    """
    check_error_on_response(response)

    element = ElementTree.fromstring(re.sub(' xmlns="[^"]+"', '', response.text, count=1))
    if find_text_or_none(element, 'errorCd') in ['-9402', '-9404', '-9403', '-9405'] \
            or find_text_or_none(element, 'msg') in ['-9402', '-9404', '-9403', '-9405']:
        raise Throttled('홈택스 이중 로그인 방지로 인해 인증에 실패했습니다. 다시 시도해주세요.')

    return element


def ensure_xml_response(fn, *args, **kwargs):
    try:
        return parse_response(fn(*args, **kwargs))
    except Throttled as e:
        if e.wait:
            time.sleep(e.wait)
            return parse_response(fn(*args, **kwargs))
        else:
            raise


def find_text_or_none(element, tag_name):
    found = element.find('.//' + tag_name)
    if found is not None:
        return found.text
    return None


class CustomHttpAdapter(requests.adapters.HTTPAdapter):
    # "Transport adapter" that allows us to use custom ssl_context.

    def __init__(self, ssl_context=None, **kwargs):
        self.ssl_context = ssl_context
        super().__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = urllib3.poolmanager.PoolManager(
            num_pools=connections, maxsize=maxsize,
            block=block, ssl_context=self.ssl_context)


class XMLValueFinder:
    element: ElementTree.Element

    def __init__(self, element):
        self.element = element

    def get(self, path, default=None):
        found = self.element.find(path)
        if found is not None:
            return found.text

        return default

    def sub_finders(self, path):
        return (XMLValueFinder(e) for e in self.element.findall(path))


def get_quarter_by_date(date):
    return int((date.month - 1) / 3 + 1)


def json_minified_dumps(obj):
    return json.dumps(obj, separators=(',', ':'), ensure_ascii=False)


import base64
import hashlib
import hmac
import re
import typing
from contextlib import contextmanager, ExitStack
from datetime import date, datetime
from typing import List

from OpenSSL import crypto
from pypinksign import pypinksign

from hometaxbot import InvalidCertificate, HometaxException


@contextmanager
def open_files(files: List[str], mode: str = 'rb') -> List[typing.BinaryIO]:
    with ExitStack() as stack:
        yield [stack.enter_context(open(f, mode)) for f in files]


def load_cert(files: List[typing.BinaryIO], password: str) -> pypinksign.PinkSign:
    for f in files:
        f.seek(0)

    ders = [f for f in reversed(files) if f.name.endswith('.der')]
    keys = [f for f in reversed(files) if f.name.endswith('.key')]
    if len(files) == 1 and (files[0].name.endswith('.pfx') or files[0].name.endswith('.p12')):
        try:
            sign = pypinksign.PinkSign(p12_data=files[0].read(),
                                       prikey_password=password.encode('utf8'))
            return sign
        except (ValueError, crypto.Error) as e:
            raise InvalidCertificate('공인인증서 비밀번호를 확인해주세요.') from e
    elif ders and keys:
        found = False
        for der in ders:
            sign = pypinksign.PinkSign(pubkey_data=der.read())
            for key in keys:
                key.seek(0)
                try:
                    sign.load_prikey(prikey_data=key.read(), prikey_password=password.encode('utf-8'))
                except ValueError:
                    continue
                if not getattr(sign, '_rand_num', None):
                    continue
                signed = sign.sign(password.encode('utf-8'))
                if sign.verify(signed, password.encode('utf-8')):
                    found = True
                    return sign
        if not found:
            raise InvalidCertificate('유효한 공인인증서를 찾을 수 없습니다. 비밀번호를 확인해주세요.')
    else:
        raise InvalidCertificate('올바른 공인인증서 형식이 아닙니다. *.p12이나 *.pfx, 또는 *.der과 *.key의 묶음 중에 하나여야 합니다.')


def validate_cert_expiry(sign):
    if sign.valid_date()[1].date() < date.today():
        raise HometaxException('공인인증서 유효기간이 지났습니다.')


def nts_hash_param(payload):
    # payload를 정확히 encode하지 않아도 API가 동작함. 그래서 형식만 맞추도록 일부만 구현함
    # 필요 시 Hometax script를 보고 정확히 해싱하여 보내도록 수정
    second = datetime.now().second
    encodedParam = base64.b64encode(hashlib.sha256((payload + str(second)).encode('utf-8')).hexdigest().upper().encode('utf-8')).decode('utf-8')
    return str(second + 12) + encodedParam + str(second)


# 홈택스 js소스에 구현된 k1~k8암호화소스


def k1(second):
    # TODO testVal 매번 바뀌는지 확인필요
    testVal = ["bakfuRUXvh9c3POvkdfUDHF91jijBhV2BvsuWE966SY", "rns6HuMkhT3FN8cIELHqYW51xHpk4oGOTetFjZ3Wog",
               "ZobgiyO5GpHBj4XfBxpIsdtuxOVGOuxfvJ3cl7hg", "tQpnppnLO4DhApYt4Wpi2fP3ikontfDj5e4gL8fatL0",
               "qyDYuOUwZO2GCykWTJJZrgRIGTg6z3FPBrIAyHxxI", "RF899ggdKY31TR3beawC7r7QbLAW1of4OrRaSWypA",
               "ASNbDSqkpdq6ckOpIoGUyO5E6xeVulnMBIQJwOAvEI"]
    return testVal[int(second) % 7]


def k2(payload, testVal):
    signature = hmac.new(testVal.encode('utf-8'), payload.encode('utf-8'), hashlib.sha256).digest()
    return re.sub(r'[^0-9a-zA-Z]', r'', base64.b64encode(signature).decode('utf-8'))


def k4(payload, second, userId=''):
    return k2(payload+userId, k1(second))


def k7():
    return f'{datetime.now().second:02}'


def k8(action_id, second):
    return k4(action_id, second)


def snake_oil_encrypt(pswd: str) -> str:
    temp_string = ""
    for ch in pswd:
        encoded_char = chr(128 - ord(ch))
        temp_string += encoded_char
    encoded_bytes = temp_string.encode("latin1")  # 0~255 범위 안전하게 인코딩
    base64_encoded = base64.b64encode(encoded_bytes).decode("ascii")
    return base64_encoded



from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum
from typing import get_type_hints
from zoneinfo import ZoneInfo

import dateutil.parser


@dataclass
class HometaxModel:
    def __setattr__(self, key, value):
        super().__setattr__(key, convert_field_type(get_type_hints(self.__class__)[key], value))


def convert_field_type(to_type, value):
    if value is None:
        return value

    if (to_type, type(value)) in hometax_field_mapping:
        return hometax_field_mapping[to_type, type(value)](to_type, value)

    if isinstance(to_type, type) and not isinstance(value, to_type):
        return to_type(value)

    return value


hometax_field_mapping = {
    (date, str): lambda to_type, value: parse_date(value),
    (datetime, str): lambda to_type, value: dateutil.parser.parse(value).replace(tzinfo=ZoneInfo('Asia/Seoul')),
    (Enum, str): lambda to_type, value: to_type(value),
    (Enum, Enum): lambda to_type, value: value,
}


def parse_date(value: str) -> date:
    if len(value) == 6:
        value += '01'
    return dateutil.parser.parse(value).date()